"""J.A.R.V.I.S. BanCog."""
import re
from datetime import datetime, timedelta

from ButtonPaginator import Paginator
from discord import User
from discord.ext import commands
from discord.utils import find
from discord_slash import SlashContext, cog_ext
from discord_slash.model import ButtonStyle
from discord_slash.utils.manage_commands import create_choice, create_option

from jarvis.db.models import Ban, Unban
from jarvis.utils import build_embed
from jarvis.utils.cachecog import CacheCog
from jarvis.utils.field import Field
from jarvis.utils.permissions import admin_or_permissions


class BanCog(CacheCog):
    """J.A.R.V.I.S. BanCog."""

    def __init__(self, bot: commands.Bot):
        super().__init__(bot)

    async def discord_apply_ban(
        self,
        ctx: SlashContext,
        reason: str,
        user: User,
        duration: int,
        active: bool,
        fields: list,
        mtype: str,
    ) -> None:
        """Apply a Discord ban."""
        await ctx.guild.ban(user, reason=reason)
        _ = Ban(
            user=user.id,
            username=user.name,
            discrim=user.discriminator,
            reason=reason,
            admin=ctx.author.id,
            guild=ctx.guild.id,
            type=mtype,
            duration=duration,
            active=active,
        ).save()

        embed = build_embed(
            title="User Banned",
            description=f"Reason: {reason}",
            fields=fields,
        )

        embed.set_author(
            name=user.nick if user.nick else user.name,
            icon_url=user.avatar_url,
        )
        embed.set_thumbnail(url=user.avatar_url)
        embed.set_footer(text=f"{user.name}#{user.discriminator} | {user.id}")

        await ctx.send(embed=embed)

    async def discord_apply_unban(self, ctx: SlashContext, user: User, reason: str) -> None:
        """Apply a Discord unban."""
        await ctx.guild.unban(user, reason=reason)
        _ = Unban(
            user=user.id,
            username=user.name,
            discrim=user.discriminator,
            guild=ctx.guild.id,
            admin=ctx.author.id,
            reason=reason,
        ).save()

        embed = build_embed(
            title="User Unbanned",
            description=f"<@{user.id}> was unbanned",
            fields=[Field(name="Reason", value=reason)],
        )
        embed.set_author(
            name=user.name,
            icon_url=user.avatar_url,
        )
        embed.set_thumbnail(url=user.avatar_url)
        embed.set_footer(text=f"{user.name}#{user.discriminator} | {user.id}")
        await ctx.send(embed=embed)

    @cog_ext.cog_slash(
        name="ban",
        description="Ban a user",
        options=[
            create_option(
                name="user",
                description="User to ban",
                option_type=6,
                required=True,
            ),
            create_option(
                name="reason",
                description="Ban reason",
                required=True,
                option_type=3,
            ),
            create_option(
                name="btype",
                description="Ban type",
                option_type=3,
                required=False,
                choices=[
                    create_choice(value="perm", name="Permanent"),
                    create_choice(value="temp", name="Temporary"),
                    create_choice(value="soft", name="Soft"),
                ],
            ),
            create_option(
                name="duration",
                description="Ban duration in hours if temporary",
                required=False,
                option_type=4,
            ),
        ],
    )
    @admin_or_permissions(ban_members=True)
    async def _ban(
        self,
        ctx: SlashContext,
        user: User = None,
        reason: str = None,
        btype: str = "perm",
        duration: int = 4,
    ) -> None:
        if not user or user == ctx.author:
            await ctx.send("You cannot ban yourself.", hidden=True)
            return
        if user == self.bot.user:
            await ctx.send("I'm afraid I can't let you do that", hidden=True)
            return
        if btype == "temp" and duration < 0:
            await ctx.send("You cannot set a temp ban to < 0 hours.", hidden=True)
            return
        elif btype == "temp" and duration > 744:
            await ctx.send("You cannot set a temp ban to > 1 month", hidden=True)
            return
        if len(reason) > 100:
            await ctx.send("Reason must be < 100 characters", hidden=True)
            return
        if not reason:
            reason = "Mr. Stark is displeased with your presence. Please leave."

        await ctx.defer()

        mtype = btype
        if mtype == "perm":
            mtype = "perma"

        guild_name = ctx.guild.name
        user_message = f"You have been {mtype}banned from {guild_name}." + f" Reason:\n{reason}"
        if mtype == "temp":
            user_message += f"\nDuration: {duration} hours"

        fields = [Field(name="Type", value=mtype)]

        if mtype == "temp":
            fields.append(Field(name="Duration", value=f"{duration} hour(s)"))

        user_embed = build_embed(
            title=f"You have been banned from {ctx.guild.name}",
            description=f"Reason: {reason}",
            fields=fields,
        )

        user_embed.set_author(
            name=ctx.author.name + "#" + ctx.author.discriminator,
            icon_url=ctx.author.avatar_url,
        )
        user_embed.set_thumbnail(url=ctx.guild.icon_url)

        try:
            await user.send(embed=user_embed)
        except Exception:
            send_failed = True
        try:
            await ctx.guild.ban(user, reason=reason)
        except Exception as e:
            await ctx.send(f"Failed to ban user:\n```\n{e}\n```", hidden=True)
            return
        send_failed = False
        if mtype == "soft":
            await ctx.guild.unban(user, reason="Ban was softban")

        fields.append(Field(name="DM Sent?", value=str(not send_failed)))
        if btype != "temp":
            duration = None
        active = True
        if btype == "soft":
            active = False

        await self.discord_apply_ban(ctx, reason, user, duration, active, fields, mtype)

    @cog_ext.cog_slash(
        name="unban",
        description="Unban a user",
        options=[
            create_option(
                name="user",
                description="User to unban",
                option_type=3,
                required=True,
            ),
            create_option(
                name="reason",
                description="Unban reason",
                required=True,
                option_type=3,
            ),
        ],
    )
    @admin_or_permissions(ban_members=True)
    async def _unban(
        self,
        ctx: SlashContext,
        user: str,
        reason: str,
    ) -> None:
        if len(reason) > 100:
            await ctx.send("Reason must be < 100 characters", hidden=True)
            return

        orig_user = user
        discrim = None
        discord_ban_info = None
        database_ban_info = None

        bans = await ctx.guild.bans()

        # Try to get ban information out of Discord
        if re.match("^[0-9]{1,}$", user):  # User ID
            user = int(user)
            discord_ban_info = find(lambda x: x.user.id == user, bans)
        else:  # User name
            if re.match("#[0-9]{4}$", user):  # User name has discrim
                user, discrim = user.split("#")
            if discrim:
                discord_ban_info = find(
                    lambda x: x.user.name == user and x.user.discriminator == discrim,
                    bans,
                )
            else:
                results = [x for x in filter(lambda x: x.user.name == user, bans)]
                if results:
                    if len(results) > 1:
                        active_bans = []
                        for ban in bans:
                            active_bans.append("{0} ({1}): {2}".format(ban.user.name, ban.user.id, ban.reason))
                        ab_message = "\n".join(active_bans)
                        message = f"More than one result. Please use one of the following IDs:\n```{ab_message}\n```"
                        await ctx.send(message)
                        return
                    else:
                        discord_ban_info = results[0]

        # If we don't have the ban information in Discord,
        # try to find the relevant information in the database.
        # We take advantage of the previous checks to save CPU cycles
        if not discord_ban_info:
            if isinstance(user, int):
                database_ban_info = Ban.objects(guild=ctx.guild.id, user=user, active=True).first()
            else:
                search = {
                    "guild": ctx.guild.id,
                    "username": user,
                    "active": True,
                }
                if discrim:
                    search["discrim"] = discrim
                database_ban_info = Ban.objects(**search).first()

        if not discord_ban_info and not database_ban_info:
            await ctx.send(f"Unable to find user {orig_user}", hidden=True)

        elif discord_ban_info:
            await self.discord_apply_unban(ctx, discord_ban_info.user, reason)
        else:
            discord_ban_info = find(lambda x: x.user.id == database_ban_info["id"], bans)
            if discord_ban_info:
                await self.discord_apply_unban(ctx, discord_ban_info.user, reason)
            else:
                database_ban_info.active = False
                database_ban_info.save()
                _ = Unban(
                    user=database_ban_info.user,
                    username=database_ban_info.username,
                    discrim=database_ban_info.discrim,
                    guild=ctx.guild.id,
                    admin=ctx.author.id,
                    reason=reason,
                ).save()
                await ctx.send("Unable to find user in Discord, " + "but removed entry from database.")

    @cog_ext.cog_subcommand(
        base="bans",
        name="list",
        description="List bans",
        options=[
            create_option(
                name="type",
                description="Ban type",
                option_type=4,
                required=False,
                choices=[
                    create_choice(value=0, name="All"),
                    create_choice(value=1, name="Permanent"),
                    create_choice(value=2, name="Temporary"),
                    create_choice(value=3, name="Soft"),
                ],
            ),
            create_option(
                name="active",
                description="Active bans",
                option_type=4,
                required=False,
                choices=[
                    create_choice(value=1, name="Yes"),
                    create_choice(value=0, name="No"),
                ],
            ),
        ],
    )
    @admin_or_permissions(ban_members=True)
    async def _bans_list(self, ctx: SlashContext, type: int = 0, active: int = 1) -> None:
        active = bool(active)
        exists = self.check_cache(ctx, type=type, active=active)
        if exists:
            await ctx.defer(hidden=True)
            await ctx.send(
                f"Please use existing interaction: {exists['paginator']._message.jump_url}",
                hidden=True,
            )
            return
        types = [0, "perm", "temp", "soft"]
        search = {"guild": ctx.guild.id}
        if active:
            search["active"] = True
        if type > 0:
            search["type"] = types[type]
        bans = Ban.objects(**search).order_by("-created_at")
        db_bans = []
        fields = []
        for ban in bans:
            if not ban.username:
                user = await self.bot.fetch_user(ban.user)
                ban.username = user.name if user else "[deleted user]"
            fields.append(
                Field(
                    name=f"Username: {ban.username}#{ban.discrim}",
                    value=(
                        f"Date: {ban.created_at.strftime('%d-%m-%Y')}\n"
                        f"User ID: {ban.user}\n"
                        f"Reason: {ban.reason}\n"
                        f"Type: {ban.type}\n\u200b"
                    ),
                    inline=False,
                )
            )
            db_bans.append(ban.user)
        if type == 0 and active:
            bans = await ctx.guild.bans()
            for ban in bans:
                if ban.user.id not in db_bans:
                    fields.append(
                        Field(
                            name=f"Username: {ban.user.name}#" + f"{ban.user.discriminator}",
                            value=(
                                f"Date: [unknown]\n"
                                f"User ID: {ban.user.id}\n"
                                f"Reason: {ban.reason}\n"
                                "Type: manual\n\u200b"
                            ),
                            inline=False,
                        )
                    )

        pages = []
        title = "Active " if active else "Inactive "
        if type > 0:
            title += types[type]
            if type == 1:
                title += "a"
        title += "bans"
        if len(fields) == 0:
            embed = build_embed(
                title=title,
                description=f"No {'in' if not active else ''}active bans",
                fields=[],
            )
            embed.set_thumbnail(url=ctx.guild.icon_url)
            pages.append(embed)
        else:
            for i in range(0, len(bans), 5):
                embed = build_embed(title=title, description="", fields=fields[i : i + 5])  # noqa: E203
                embed.set_thumbnail(url=ctx.guild.icon_url)
                pages.append(embed)

        paginator = Paginator(
            bot=self.bot,
            ctx=ctx,
            embeds=pages,
            only=ctx.author,
            timeout=60 * 5,  # 5 minute timeout
            disable_after_timeout=True,
            use_extend=len(pages) > 2,
            left_button_style=ButtonStyle.grey,
            right_button_style=ButtonStyle.grey,
            basic_buttons=["◀", "▶"],
        )

        self.cache[hash(paginator)] = {
            "guild": ctx.guild.id,
            "user": ctx.author.id,
            "timeout": datetime.utcnow() + timedelta(minutes=5),
            "command": ctx.subcommand_name,
            "type": type,
            "active": active,
            "paginator": paginator,
        }

        await paginator.start()
